#include <iostream>
#include <string>
#include <vector>
#include <utility>
#include <map>
#include <ctime>

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/objdetect/objdetect.hpp>
#include <opencv2/ml/ml.hpp>

#include "Settings.h"
#include "FaceProcessing.h"
#include "GaborFilter.h"
#include "Classifier.h"

using namespace std;
using namespace cv;

int main(int argc, const char* argv[]) {

	namedWindow("Facepression", 1);

	VideoCapture captureDevice;
	captureDevice.open(0);

	CascadeClassifier faceCascade;
	faceCascade.load(faceCascadeFile);
	
	// vektor realnych a imaginarnich kernelu pro gaboruv filtr
	vector<pair<Mat, Mat>> gaborBank;
	fillGaborBank(gaborBank);

	// klasifikatory vyrazu
	vector<pair<string, CvSVM*>> classifiers;
	createClassifiers(classifiers);

	int key;

	while (1) {
		Mat frame, face, filtered, features;

		captureDevice >> frame;

		clock_t tStart = clock();

		if (findFace(frame, face, faceCascade)) {
			gaborFilter(face, filtered, gaborBank);
			features = filtered.reshape(1, 1);

			pair<string, float> predictedExpression = predictExpression(classifiers, features);

			string expression = predictedExpression.first;
			int lineLength = (int)(-200 * predictedExpression.second);

			putText(frame, expression.c_str(), Point(20, 40), FONT_HERSHEY_SIMPLEX, 0.6, Scalar(255, 255, 255));
			line(frame, Point(21, 50), Point(21 + lineLength, 50), Scalar(255, 255, 255), 1);
		}
		
		imshow("Facepression", frame);

		printf("time: %.3fs\n", (double)(clock() - tStart) / CLOCKS_PER_SEC);

		key = waitKey(10);
		if (key == 27) {
			break;
		}
		else if (key == 115) {
			string screenshot = dataDir;
			screenshot.append(screenshotsSubDir).append(to_string(time(nullptr))).append(".png");
			imwrite(screenshot, frame);
		}
	}

	return 0;
}
