#include <iostream>
#include <fstream>
#include <ctime>

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/objdetect/objdetect.hpp>
#include <opencv2/ml/ml.hpp>

#include "Settings.h"
#include "FaceProcessing.h"
#include "GaborFilter.h"
#include "Classifier.h"
#include "Actions.h"

using namespace std;
using namespace cv;

void cropFaces() {
	clock_t tStart = clock();

	CascadeClassifier faceCascade;
	faceCascade.load(faceCascadeFile);

	string inputDir = dataDir;
	inputDir.append(originalSubDir);
	string outputDir = dataDir;
	outputDir.append(facesSubDir);

	string inputImgList;
	string inputImgName;
	string inputImgPath;

	string outputImgPath;

	for (vector<string>::iterator it = expressions.begin(); it != expressions.end(); ++it) {
		string expression = *it;
		cout << "[" << expression << "]" << endl;
		inputImgList = inputDir;
		inputImgList.append(expression).append("/list.txt");
		ifstream imgListFile(inputImgList);
		while (imgListFile >> inputImgName)
		{
			cout << "Zpracovavam " << inputImgName << " ... ";

			inputImgPath.clear();
			inputImgPath.append(inputDir).append(expression).append("/").append(inputImgName).append("a");

			inputImgName = inputImgName.substr(0, inputImgName.find_last_of("."));
			outputImgPath.clear();
			outputImgPath.append(outputDir).append(expression).append("/").append(inputImgName).append(".png");

			Mat frame, face;
			frame = imread(inputImgPath, 1);
			if (frame.empty()) {
				cout << "soubor neexistuje" << endl;
				continue;
			}

			bool result;
			result = findFace(frame, face, faceCascade);
			if (result) {
				imwrite(outputImgPath, face);
			}

			cout << "hotovo (" << result << " oblicej)" << endl;
		}
		imgListFile.close();
		cout << endl;
	}

	printf("Cas: %.3fs\n", (double)(clock() - tStart) / CLOCKS_PER_SEC);
}

void extractFeatures() {
	clock_t tStart = clock();

	vector<pair<Mat, Mat>> gaborBank;
	fillGaborBank(gaborBank);

	string originalDir = dataDir;
	originalDir.append(originalSubDir);
	string inputDir = dataDir;
	inputDir.append(facesSubDir);
	string outputDir = dataDir;
	outputDir.append(featuresSubDir);

	string inputImgList;
	string inputImgName;
	string inputImgPath;

	string outputImgPath;

	for (vector<string>::iterator it = expressions.begin(); it != expressions.end(); ++it) {
		string expression = *it;
		cout << "[" << expression << "]" << endl;
		inputImgList = originalDir;
		inputImgList.append(expression).append("/list.txt");
		ifstream imgListFile(inputImgList);
		while (imgListFile >> inputImgName)
		{
			inputImgName = inputImgName.substr(0, inputImgName.find_last_of("."));
			cout << "Zpracovavam " << inputImgName << " ... ";

			inputImgPath.clear();
			inputImgPath.append(inputDir).append(expression).append("/").append(inputImgName).append(".png");
			outputImgPath.clear();
			outputImgPath.append(outputDir).append(expression).append("/").append(inputImgName).append(".yml");

			Mat face, filtered;
			face = imread(inputImgPath, CV_LOAD_IMAGE_GRAYSCALE);
			if (face.empty()) {
				cout << "soubor neexistuje" << endl;
				continue;
			}

			gaborFilter(face, filtered, gaborBank);
			FileStorage fs(outputImgPath, FileStorage::WRITE);
			fs << inputImgName << filtered;
			fs.release();

			cout << "hotovo" << endl;
		}
		imgListFile.close();
		cout << endl;
	}

	printf("Cas: %.3fs\n", (double)(clock() - tStart) / CLOCKS_PER_SEC);
}

void prepareTrainingSets() {
	clock_t tStart = clock();

	string originalDir = dataDir;
	originalDir.append(originalSubDir);
	string outputDir = dataDir;
	outputDir.append(trainingSubDir);

	string inputImgList;
	string inputImgName;

	string outputFilePath;
	for (vector<string>::iterator it = expressions.begin(); it != expressions.end(); ++it) {
		string expression = *it;

		cout << "[" << expression << "]" << endl;

		inputImgList = originalDir;
		inputImgList.append(expression).append("/list.txt");
		outputFilePath = outputDir;
		outputFilePath.append(expression).append(".csv");

		ofstream outputFile(outputFilePath);

		for (vector<string>::iterator it2 = expressions.begin(); it2 != expressions.end(); ++it2) {
			inputImgList.clear();
			inputImgList.append(originalDir).append(*it2).append("/list.txt");
			ifstream imgListFile(inputImgList);
			while (imgListFile >> inputImgName)
			{
				inputImgName = inputImgName.substr(0, inputImgName.find_last_of("."));
				cout << "Zpracovavam " << inputImgName << " ... ";

				inputImgName.append(".yml");

				if (*it2 == expression) {
					outputFile << inputImgName << ",1," << *it2 << endl;
				}
				else {
					outputFile << inputImgName << ",0," << *it2 << endl;
				}

				cout << "hotovo" << endl;
			}
			imgListFile.close();
		}

		outputFile.close();
		cout << endl;
	}

	printf("Cas: %.3fs\n", (double)(clock() - tStart) / CLOCKS_PER_SEC);
}

void train() {
	clock_t tStart = clock();

	string featuresDir = dataDir;
	featuresDir.append(featuresSubDir);
	string trainingDir = dataDir;
	trainingDir.append(trainingSubDir);
	string classifiersDir = dataDir;
	classifiersDir.append(classifiersSubDir);

	string outputFilePath;

	// ziskani rozmeru priznaku z prvniho souboru v prvnim csv
	string trainingFilePath1 = trainingDir;
	trainingFilePath1.append(expressions.front()).append(".csv");
	string sampleFile1;
	string sampleFile1Path;
	int featuresSize;

	ifstream trainingFile1(trainingFilePath1);
	if (getline(trainingFile1, sampleFile1, ',')) {
		sampleFile1 = sampleFile1.substr(0, sampleFile1.find_last_of("."));
		sampleFile1Path.clear();
		sampleFile1Path.append(featuresDir).append(expressions.front()).append("/").append(sampleFile1).append(".yml");

		Mat sampleFileData1;
		FileStorage fs1(sampleFile1Path, FileStorage::READ);
		fs1[sampleFile1] >> sampleFileData1;
		fs1.release();

		featuresSize = sampleFileData1.rows * sampleFileData1.cols;
	}
	else {
		cout << "Chyba: soubor s priznaky " << sampleFile1Path << " neexistuje" << endl;
		return;
	}
	trainingFile1.close();

	Mat trainingData(0, featuresSize, CV_32FC1);

	// pres vsechny vyrazy
	for (vector<string>::iterator it = expressions.begin(); it != expressions.end(); ++it) {
		string expression = *it;

		cout << "[" << expression << "]" << endl;

		string trainingFilePath = trainingDir;
		trainingFilePath.append(expression).append(".csv");

		string sampleFile;
		string sampleFilePath;
		string sampleClassString;
		int sampleClass;
		string sampleExpression;

		Mat labels(0, 1, CV_32FC1);

		ifstream trainingFile(trainingFilePath);
		while (getline(trainingFile, sampleFile, ',')) {
			getline(trainingFile, sampleClassString, ',');
			getline(trainingFile, sampleExpression, '\n');

			sampleClass = stoi(sampleClassString);

			if (expression == expressions.front()) {
				sampleFile = sampleFile.substr(0, sampleFile.find_last_of("."));

				sampleFilePath.clear();
				sampleFilePath.append(featuresDir).append(sampleExpression).append("/").append(sampleFile).append(".yml");

				cout << "Nacitam priznaky " << sampleFile << " (" << sampleExpression << ") ... ";

				Mat sampleFileData;
				FileStorage fs(sampleFilePath, FileStorage::READ);
				fs[sampleFile] >> sampleFileData;
				fs.release();

				trainingData.push_back(sampleFileData.reshape(1, 1));

				cout << "hotovo" << endl;
			}
			else {
				cout << ".";
			}

			labels.push_back(sampleClass);

		}
		trainingFile.close();

		cout << endl;
		printf("Cas: %.3fs\n", (double)(clock() - tStart) / CLOCKS_PER_SEC); // po nacteni vsech trenovacich souboru
		cout << "Trenuji klasifikator [" << expression << "] ... ";

		// trenovani klasifikatoru
		string classifiersFilePath = classifiersDir;
		classifiersFilePath.append(expression).append(".xml");

		CvSVM svm;
		CvSVMParams params;

		params.svm_type = CvSVM::C_SVC;
		params.kernel_type = CvSVM::LINEAR;
		params.C = SVMCvalue;
		params.term_crit = cvTermCriteria(CV_TERMCRIT_ITER, SVMMaxIteration, SVMEpsilon);

		svm.train(trainingData, labels, Mat(), Mat(), params);
		svm.save(classifiersFilePath.c_str());

		cout << "hotovo" << endl;

		printf("Cas: %.3fs\n", (double)(clock() - tStart) / CLOCKS_PER_SEC);
	}

	printf("Cas: %.3fs\n", (double)(clock() - tStart) / CLOCKS_PER_SEC);
}

void test() {
	clock_t tStart = clock();

	string inputDir = dataDir;
	inputDir.append(testSubDir);
	string outputFilePath = dataDir;
	outputFilePath.append(testSubDir).append("results.csv");

	ofstream outputFile(outputFilePath);

	string inputImgList;
	string inputImgName;
	string inputImgPath;

	CascadeClassifier faceCascade;
	faceCascade.load(faceCascadeFile);

	// vektor realnych a imaginarnich kernelu pro gaboruv filtr
	vector<pair<Mat, Mat>> gaborBank;
	fillGaborBank(gaborBank);

	// klasifikatory vyrazu
	vector<pair<string, CvSVM*>> classifiers;
	createClassifiers(classifiers);

	for (vector<string>::iterator it = expressions.begin(); it != expressions.end(); ++it) {
		string expression = *it;
		cout << "[" << expression << "]" << endl;
		inputImgList = inputDir;
		inputImgList.append(expression).append("/list.txt");
		ifstream imgListFile(inputImgList);
		while (imgListFile >> inputImgName)
		{
			cout << "Testuji " << inputImgName << " - " << expression << " ... ";

			inputImgPath.clear();
			inputImgPath.append(inputDir).append(expression).append("/").append(inputImgName);

			Mat frame, face, filtered, features;

			frame = imread(inputImgPath, 1);
			if (frame.empty()) {
				cout << "soubor neexistuje" << endl;
				continue;
			}

			if (findFace(frame, face, faceCascade)) {
				gaborFilter(face, filtered, gaborBank);
				features = filtered.reshape(1, 1);

				pair<string, float> predictedExpression = predictExpression(classifiers, features);

				// csv:       nazev souboru          skutecny vyraz       klasifikovany vyraz                 vzdalenost od hyperroviny
				outputFile << inputImgName << "," << expression << "," << predictedExpression.first << "," << predictedExpression.second << endl;
				cout << predictedExpression.first << "?" << endl;
			}
			else {
				cout << "neobsahuje oblicej" << endl;
			}
		}
		imgListFile.close();
		cout << endl;
	}

	outputFile.close();

	printf("Cas: %.3fs\n", (double)(clock() - tStart) / CLOCKS_PER_SEC);
}