#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <utility>

#include <opencv2/core/core.hpp>
#include <opencv2/ml/ml.hpp>

#include "Settings.h"
#include "Classifier.h"

using namespace std;
using namespace cv;

void createClassifiers(vector<pair<string, CvSVM*>>& classifiers) {
	for (vector<string>::iterator it = expressions.begin(); it != expressions.end(); ++it) {
		string expression = *it;

		string classifierPath = dataDir;
		classifierPath.append(classifiersSubDir).append(expression).append(".xml");

		CvSVM* svm = new CvSVM;
		svm->load(classifierPath.c_str());
		classifiers.push_back(pair<string, CvSVM*>(expression, svm));
	}
}

pair<string, float> predictExpression(vector<pair<string, CvSVM*>>& classifiers, Mat& features) {
	vector<pair<string, float>> results;

	for (vector<pair<string, CvSVM*>>::iterator it = classifiers.begin(); it != classifiers.end(); ++it) {
		string expression = it->first;
		CvSVM* svm = it->second;

		float prediction = svm->predict(features, true);
		
		results.push_back(make_pair(expression, prediction));
	}

	pair<string, float> best = make_pair("unknown", 0);

	for (vector<pair<string, float>>::iterator it2 = results.begin(); it2 != results.end(); ++it2) {
		if (it2->second < best.second) {
			best.first = it2->first;
			best.second = it2->second;
		}
	}

	return best;
}

